import base64
import re
import time
from typing import Type

import requests

from .types import File, FileUploader, MultipartProgress, ProgressHandler


class HostManager:
    def __init__(self):
        self.hosts = []

    def register_host(self, host: Type[FileUploader]):
        self.hosts.append(host)
        return host


host_manager = HostManager()


@host_manager.register_host
class MirroredToUploader(FileUploader):
    site = "mirrored.to"
    short = "mir"

    def __init__(self, file: File, progress: ProgressHandler):
        self.file = file
        self.progress = progress
        self.re_mir_token = re.compile(r"'token'     : '([a-z0-9]{32})'")
        self.re_mir_files = re.compile(
            r"(https://www\.mirrored\.to/files/[A-Z0-9]+/[A-z0-9._]+)<"
        )
        self.re_mir_hfiles = re.compile(
            r"https://www\.mirrored\.to/files/[A-Z0-9]+/\?hash=[0-9a-f]+&dl=[01]"
        )
        self.re_mir_mirstats = re.compile(
            r"/mirstats\.php\?uid=[A-Z0-9]+&tmpID=[0-9a-f]+&fn=[A-z0-9.]+&ads=1&gp=1&su=0&pid=0&puid=0&fd=1&s=0&lang=[a-z]+&ftype=[A-z]+"
        )
        self.re_mir_getlink = re.compile(
            r"/getlink/[A-Z0-9]+/[0-9]+/\?hid=[A-z0-9%]+&tid=[0-9a-f]+"
        )
        self.re_mir_hosts = re.compile(
            r"f=(https?:\/\/(?:www\.)?[-a-zA-Z0-9@:%._\+~#=]{1,256}\.[a-zA-Z0-9()]{1,6}\b(?:[-a-zA-Z0-9()@:%_\+.~#?&//=]*)) target=\"_blank\">"
        )

    def upload(self) -> None:
        self.progress.make_visible()
        try:
            self._upload()
            self.progress.done()
        except Exception as e:
            self.progress.exception(f"Failed uploading to {self.site} due to: {e}")
            self.progress.failed()

    def _upload(self):
        self.progress.update_description("Fetching token...")
        r = requests.get("https://www.mirrored.to/")
        rg = self.re_mir_token.search(r.text)
        if rg is None:
            self.progress.exception("No token found!")
            return
        token = rg.group(1)

        self.progress.update_description("Uploading...", True)
        r = requests.post(
            "https://www.mirrored.to/uploadify/uploadifive1.php",
            **MultipartProgress(
                self.progress,
                {
                    "timestamp": "",
                    "token": token,
                    "Filedata": (self.file.name, self.file),
                },
            ),
        )
        if not r.ok and r.text:
            self.progress.exception("Failed to upload")
            return

        for hosts in (
            (
                "gofileio",
                "downloadgg",
                "onefichier",
                "turbobit",
                "zippyshare",
                "usersdrive",
                "bayfiles",
                "anonfiles",
                "clicknupload",
                "uptobox",
            ),
            (
                "dailyuploads",
                "uploadee",
                "dropapk",
                "mixdropco",
                "filesim",
                "megaupnet",
                "file-upload",
                "sendcm",
                "skynet",
                "pixeldrain",
            ),
        ):
            # B64 ENCODE:
            # For each filename:
            #    Enter name of file. Be sure to use format of filename listed in upload result as guide. It must be exact. Do not include quotes.
            #    Paste this: #0#
            #    Enter size of file in Bytes. For 1 MiB file, enter 1048576. This must be exact!
            #    Paste this: ;0;
            # Example for 2 files:
            # First_File.7z#0#1234567;0;Second_File.7z#0#8901234;0;
            # Paste this: @e@#H#
            # Paste host list from box below.

            # onefichier;anonfiles;solidfiles
            # Paste this: ;#P##SC##T#
            # Enter some numbers. Example: 1625815023. It should be unique, as it is used for timestamp.

            self.progress.update_description("Fetching link page...")
            data = base64.b64encode(
                f"{self.file.name}#0#{self.file.size};0;@e@#H#{';'.join(hosts)};#P##SC##T#{int(time.time() * 1000)}".encode()
            ).decode()
            r = requests.get(
                "https://www.mirrored.to/upload_complete.php",
                params={"w": "1", "data": data},
            )
            rg = self.re_mir_files.search(r.text)
            if rg is None:
                self.progress.exception("No link to url list in upload_complete")
                return
            files_url = rg.group(1)

            r = requests.get(files_url)
            rg = self.re_mir_hfiles.search(r.text)
            if rg is None:
                self.progress.exception("No hash link to list in files url")
                return
            hfiles_url = rg.group(0)

            self.progress.update_description("Waiting for upload...")
            r = requests.get(hfiles_url)
            rg = self.re_mir_mirstats.search(r.text)
            if rg is None:
                self.progress.exception("No mirstats link in hfiles url")
                return
            mirstats_url = "https://www.mirrored.to" + rg.group(0)

            for i in range(300):  # Timeout to avoid waiting forever
                r = requests.get(mirstats_url)
                if "id_Uploading" not in r.text:
                    break
                time.sleep(5)
            getlinks = self.re_mir_getlink.findall(r.text)

            self.progress.update_description("Fetching links...")
            for link in getlinks:
                r = requests.get("https://mirrored.to/" + link)
                rg = self.re_mir_hosts.search(r.text)
                if rg is None:
                    continue  # It's not worth trashing all other mirrors
                self.progress.add_url(rg.group(1))


@host_manager.register_host
class AnonfilesUploader(FileUploader):
    site = "anonfiles.com"
    short = "an"

    def __init__(self, file: File, progress: ProgressHandler):
        self.file = file
        self.progress = progress

    def upload(self) -> None:
        self.progress.make_visible()
        try:
            self._upload()
            self.progress.done()
        except Exception as e:
            self.progress.exception(f"Failed uploading to {self.site} due to: {e}")
            self.progress.failed()

    def _upload(self) -> None:
        self.progress.update_description("Uploading...", True)
        j = requests.post(
            "https://api.anonfiles.com/upload",
            **MultipartProgress(self.progress, {"file": (self.file.name, self.file)}),
        ).json()
        self.progress.add_url(j["data"]["file"]["url"]["short"])


@host_manager.register_host
class TusfilesUploader(FileUploader):
    site = "tusfiles.com"
    short = "tus"

    def __init__(self, file: File, progress: ProgressHandler):
        self.file = file
        self.progress = progress

    def upload(self) -> None:
        self.progress.make_visible()
        try:
            self._upload()
            self.progress.done()
        except Exception as e:
            self.progress.exception(f"Failed uploading to {self.site} due to: {e}")
            self.progress.failed()

    def _upload(self) -> None:
        self.progress.update_description("Uploading...", True)
        j = requests.post(
            "https://cloud01.tusfiles.com/cgi-bin/upload.cgi",
            params={"upload_type": "file", "utype": "anon"},
            **MultipartProgress(
                self.progress,
                {
                    "file_0": (self.file.name, self.file),
                    "sess_id": "",
                    "utype": "anon",
                    "link_pass": "",
                    "link_rcpt": "",
                    "link_pass": "",
                    "keepalive": "1",
                }
            ),
        ).json()
        self.progress.add_url("https://tusfiles.com/" + j[0]["file_code"])
